# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""An abstract base class for simple, character-oriented tokenizers.
"""

from tokenizer import Tokenizer
from token import Token

IO_BUFFER_SIZE = 1024
MAX_WORD_LEN = 255

class CharTokenizer(Tokenizer):
    __slots__ = ['buffer', 'bufferIndex', 'dataLen', 'ioBuffer', 'offset']

    def __init__(self, input):
        self.buffer = [0] * MAX_WORD_LEN
        self.bufferIndex = 0
        self.offset = 0
        self.dataLen = 0
        self.ioBuffer = ''
        self.input = input


    def isTokenChar(self, c):
        """Returns true iff a character should be included in a token.  This
        * tokenizer generates as tokens adjacent sequences of characters which
        * satisfy this predicate.  Characters for which this is false are used to
        * define token boundaries and are not included in tokens."""


    def next(self):
        #Returns the next token in the stream, or None at EOS.
        length = 0
        start = self.offset
        while True:
            self.offset += 1
            if (self.bufferIndex >= self.dataLen):
                self.ioBuffer = self.input.read(IO_BUFFER_SIZE)
                self.dataLen = len(self.ioBuffer)
                self.bufferIndex = 0

            if self.dataLen == 0:
                if length > 0:
                    break
                else:
                    return None
            else:
                c = self.ioBuffer[self.bufferIndex]
                self.bufferIndex += 1

            if self.isTokenChar(c):
                # if it's a token char
                if (length == 0):
                    #start of token
                    start = self.offset-1
                # buffer it normalized
                self.buffer[length] = self.normalize(c)
                length += 1

                if (length == MAX_WORD_LEN):
                    #buffer overflow!
                    break
            elif length > 0:
                # at non-Letter w/ chars
                break

        # is there a better way of turning a list of chars into a string?
        tstr = ''.join(self.buffer[0:length])
        return Token(tstr, start, start+length)



    def normalize(self, c):
        """Called on each token character to normalize it before it is added to the
        * token.  The default implementation does nothing.  Subclasses may use this
        * to, e.g., lowercase tokens"""
        return c
