# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""A field is a section of a Document.

Each field has two parts, a name and a value.  Values may be free
text, provided as a String or as a Reader, or they may be atomic
keywords, which are not further processed.  Such keywords may be used
to represent dates, urls, etc.  Fields are optionally stored in the
index, so that they may be returned with hits on the document.
"""

import types


class Field(object):

    def __init__(self, name, string, store=False, index=True, token=True):
        self.nom = name
        self.stringVal = string
        self.readerVal = None
        self.isStored = store
        self.isIndexed = index
        self.isTokenized = token        

    def __repr__(self):
        if self.isStored and self.isIndexed and not self.isTokenized:
            return '<Keyword<' + self.nom + ':' + self.stringVal + '>>'
        elif self.isStored and not self.isIndexed and not self.isTokenized:
            return '<Unindexed<' + self.nom + ':' + self.stringVal + '>>'
        elif self.isStored and self.isIndexed and self.isTokenized and self.stringVal is not None:
            return '<Text<' + self.nom + ':' + self.stringVal + '>>'
        elif self.isStored and self.isIndexed and self.isTokenized and self.stringVal is not None:
            return '<Text<' + self.nom + ':' + self.readerVal + '>>'
        else:
            return '<Field<???>'

    def name(self):
        return self.nom

    def stringValue(self):
        return self.stringVal

    def readerValue(self):
        return self.readerVal


# XXX make these subclasses?
def Keyword(name, value):
    return Field(name, value, True, True, False)


def Text(name, strOrReader, store=True):
    """Accepts string or file-like object."""
    if isinstance(strOrReader, types.StringType):
        # Plain string
        res = Field(name, strOrReader, store, True, True)
    elif isinstance(strOrReader, types.UnicodeType):
        # Unicode string
        res = Field(name, strOrReader, store, True, True)
    else:
        # it's a reader, eg. strOrReader is a file-like object
        res = Field(name, None)
        res.readerVal = strOrReader
        res.stringVal = None
    return res


def UnIndexed(name, value):
    return Field(name, value, True, False, False)


def UnStored(name, value):
    return Field(name, value, False, True, True)


class DateField(object):
    """Provides support for converting dates to strings and vice-versa.
    
    The strings are structured so that lexicographic sorting orders by
    date. This makes them suitable for use as field values and search
    terms.
    """
    DATELEN = len(str(1000L*365*24*60*60*1000))

    def __init__(self):
        raise NotImplementedError
