# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""IndexReader is an abstract class, providing an interface for
accessing an index. Search of an index is done entirely through this abstract
interface, so that any subclass which implements it is searchable.

Concrete subclasses of IndexReader are usually constructed with a call to the
class method open(path).

For efficiency, in this API documents are often referred to via document
numbers, non-negative integers which each name a unique document in the index.
These document numbers are ephemeral--they may change as documents are added
to and deleted from an index. Clients should thus not rely on a given document
having the same number between sessions. """

import os
from lupy.store import fsdirectory
#from lupy.util import locker

class IndexReader(object):
    # class methods

    def openDir(cls, directory):
        # TODO: avoiding circular import - has to be better way
        import segmentinfos, segmentreader, segmentsreader
        # TODO Locking?
        infos = segmentinfos.SegmentInfos()
        infos.read(directory)
        if len(infos) == 1:       # index is optimized
            return segmentreader.SegmentReader(infos[0], True)

        readers = []
        for i in range(len(infos)):
            readers.append(segmentreader.SegmentReader(infos[i],
                                                       i == len(infos)-1))

        return segmentsreader.SegmentsReader(directory, readers)
    openDir = classmethod(openDir)
    

    def open(cls, path):
        """Returns an IndexReader reading the index in an FSDirectory in
        the named path."""
        
        return cls.openDir(fsdirectory.FSDirectory.getDirectory(path, False))
    open = classmethod(open)


    # instance methods

    def __init__(self, d):
        self.directory = d

    def indexExists(self, d):
        """Returns True if an index exists at the specified directory"""
        return self.directory.fileExists('segments')

    def indexExistsFile(self, d):
        return os.path.exists(os.path.join(d,'segments'))

    def indexExistsString(self, d):
        return os.path.exists(os.path.join(d,'segments'))

    def isLocked(self):
        # return self.directory.fileExists('write.lock')
        return False

    def lastModified(self, d):
        """Returns the time the index in this directory was last modified."""
        return self.directory.fileModified('segments')
    
    
    def lastModifiedFile(self, d):
        """Returns the time the index in the named directory was last modified."""
        return fsdirectory.FSDirectory.fileModified(d, 'segments')
    

    def lastModifiedString(self, d):
        return self.lastModified(d)
    

    def unlock(self, directory):
        """Forcibly unlocks the index in the named directory.
        
        Caution: this should only be used by failure recovery code,
        when it is known that no other process nor thread is in fact
        currently accessing this index."""
        
        directory.deleteFile('write.lock')
        directory.deleteFile('commit.lock')
        

    def close(self):
        """Closes files associated with this index.
        Also saves any new deletions to disk.
        No other methods should be called after this has been called."""
        
        self.doClose()
        

    def doClose(self):
        pass


    def delete(self, docNum):
        """Deletes the document numbered B{docNum}.  Once a document is
        deleted it will not appear in TermDocs or TermPostitions enumerations.
        Attempts to read its field with the L{#document}
        method will result in an error.  The presence of this document may still be
        reflected in the L{#docFreq} statistic, though
        this will be corrected eventually as the index is further modified.
        """
        self.doDelete(docNum)
        

    def deleteTerm(self, term):
        """ Deletes all documents containing B{term}.
        This is useful if one uses a document field to hold a unique ID string for
        the document.  Then to delete such a document, one merely constructs a
        term with the appropriate field and the unique ID string as its text and
        passes it to this method.  Returns the number of documents deleted.
        """
        
        docs = self.termDocsTerm(term)
        if docs is None:
            return 0

        n = 0
        try:
            while docs.next():
                self.delete(docs.doc())
                n += 1
        finally:
            docs.close()
        return n
    
    
    def termDocs(self):
        """Returns an unpositioned {@link TermDocs} enumerator.
        """
    
    
    def termDocsTerm(self, term):
        """ Returns an enumeration of all the documents which contain
        B{term}. For each document, the document number, the frequency of
        the term in that document is also provided, for use in search scoring.
        Thus, this method implements the mapping:
        <p><ul>
        Term &nbsp;&nbsp; =&gt; &nbsp;&nbsp; &lt;docNum, freq&gt;<sup>*</sup>
        </ul>
        The enumeration is ordered by document number.  Each document number
        is greater than all that precede it in the enumeration."""
        
        termDocs = self.termDocs()
        termDocs.seekTerm(term)
        return termDocs
    
    
    def termPositionsTerm(self, term):
        """Returns an enumeration of all the documents which contain
        B{term}.  For each document, in addition to the document number
        and frequency of the term in that document, a list of all of the ordinal
        positions of the term in the document is available.  Thus, this method
        implements the mapping:
    
        <p><ul>
        Term &nbsp;&nbsp; =&gt; &nbsp;&nbsp; &lt;docNum, freq,
              &lt;pos<sub>1</sub>, pos<sub>2</sub>, ...
              pos<sub>freq-1</sub>&gt;
            &gt;<sup>*</sup>
        </ul>
        <p> This positional information faciliates phrase and proximity searching.
        <p>The enumeration is ordered by document number.  Each document number is
        greater than all that precede it in the enumeration."""
        
        termPositions = self.termPositions()
        termPositions.seekTerm(term)
        return termPositions
        
            
        
        


                
        


    
