# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

import sys

from array import array

import lupy.store.ramdirectory as ramdirectory
import lupy.store.fsdirectory as fsdirectory
from indexreader import IndexReader
from termenum import TermEnum



class SegmentsReader(IndexReader):
    __slots__ = ['maxiDoc', 'normsCache', 'numiDocs', 'readers', 'starts']


    def __init__(self, directory, r):
        IndexReader.__init__(self, directory)
        self.readers = r
        self.maxiDoc = 0
        self.normsCache = {}
        self.numiDocs = -1

        self.starts = [0] * (len(self.readers) + 1)

        i = 0
        while i < len(self.readers):
            self.starts[i]=self.maxiDoc
            self.maxiDoc += self.readers[i].maxDoc()
            i += 1
        self.starts[len(self.readers)] = self.maxiDoc
            

    def docFreq(self, t):
        total = 0
        for r in self.readers:
            total += r.docFreq(t)
        return total


    def doClose(self):
        for r in self.readers:
            r.close()


    def document(self, n):
        # find segment num
        i = self.readerIndex(n)
        # dispatch to segment reader
        return self.readers[i].document(n - self.starts[i])    


    def doDelete(self, n):
        # invalidate cache
        self.numiDocs = -1
        # find seg num
        i = self.readerIndex(n)
        # dispatch to seg reader
        self.readers[i].doDelete(n - self.starts[i])


    def isDeleted(self, n):
        # find segment num
        i = self.readerIndex(n)
        # dispatch to segment reader
        return self.readers[i].isDeleted(n - self.starts[i])    


    def maxDoc(self):
        return self.maxiDoc


    def normsField(self, field):
        bytes = self.normsCache.get(field, None)
        if bytes is not None:
            # cache hit
            return bytes

        bytes = array('B',[0x00] * self.maxDoc())
        for i in range(len(self.readers)):
            self.readers[i].norms(field, bytes, self.starts[i])
        # update cache
        self.normsCache[field]=bytes
        return bytes


    def numDocs(self):
        # check cache
        if numiDocs == -1:
            # cache miss - recompute
            n = 0
            for r in self.readers:
                # sum from readers
                n += r.numDocs()
            self.numiDocs = n
        return self.numiDocs


    def readerIndex(self, n):
        # Search starts array for first element less than n
        lo = 0
        hi = len(self.readers) - 1

        while hi >= lo:
            mid = (lo + hi) >> 1
            midValue = self.starts[mid]
            if n < midValue:
                hi = mid - 1
            elif n > midValue:
                lo = mid + 1
            else:
                return mid
        return hi


    def termDocs(self):
        return SegmentsTermDocs(self.readers, self.starts)


    def termPositions(self):
        return SegmentsTermPositions(self.readers, self.starts)
    

    def terms(self):
        return SegmentsTermEnum(self.readers, self.starts)
    

    def termsOf(self, term):
        return SegmentsTermEnum(self.readers, self.starts, term)


    def fieldNames(self):
        # Experimental for auto-queries
        if self.readers:
            return self.readers[0].fieldInfos.fieldNames()
        else:
            return []



class SegmentsTermEnum(TermEnum):
    __slots__ = ['docFreq', 'queue', 'term']


    def __init__(self, readers, starts, t):
        self.queue = SegmentMergeQueue(readers.length)
        for i in range(len(readers)):
            reader = readers[i]
            
            if ti is not None:
                termEnum = reader.terms(t)
            else:
                termEnum = reader.terms()

            smi = SegmentMergeInfo(starts[i], termEnum, reader)
            if t is None:
                res = smi.next()
            else:
                res = termEnum.term() is not None

            if res:
                self.queue.put(smi)
            else:
                smi.close()

        if t is not None and len(self.queue) > 0:
            top = self.queue.top()
            self.term = top.termEnum.term()
            self.docFreq = to.termEnum.docFreq()
            
            
    def next(self):
        # Increments the enumerations to the next element. True if one exists
        top = self.queue.top()
        if top is None:
            term = None
            return False

        self.term = top.term
        self.docFreq = 0

        while(top is not None and self.term.compareTo(top.term)) == 0:
            self.queue.pop()
            # increment frequency
            self.docFreq += top.termEnum.doFreq()
            if top.next():
                # restore queue
                self.queue.put(top)
            else:
                top.close()
            top = self.queue.top
            
        return True
            

    def term(self):
        # Returns the current Term in the enumeration
        return self.term
    

    def docFreq(self):
        # Returns the docFreq of the current Term in the enumeration
        return self.docFreq
    

    def close(self):
        # Closes the enumeration to further activity, freeing resources
        self.queue.close()



class SegmentsTermDocs(object):
    __slots__ = ['base', 'current', 'pointer', 'readers',
                 'segTermDocs', 'starts', 'term']

    

    def __init__(self, r, s):
        self.readers = r
        self.starts = s

        self.base = 0
        self.pointer = 0
        self.current = None
        self.term = None
        
        self.segTermDocs = [None] * len(r)


    def close(self):
        for segtdoc in self.segTermDocs:
            if segtdoc is not None:
                segtdoc.close()


    def doc(self):
        return self.base + self.current.docu


    def freq(self):
        return self.current.frq


    def next(self):
        if self.current is not None and self.current.next():
            return True
        elif self.pointer < len(self.readers):
            self.base = self.starts[self.pointer]
            self.current = self.termDocsInt(self.pointer)
            self.pointer += 1
            return self.next()
        else:
            return False


    def read(self, docs, freqs):
        while True:
            while self.current is None:
                if self.pointer < len(self.readers):
                    # try next segment
                    self.base = self.starts[self.pointer]
                    self.current = self.termDocsInt(self.pointer)
                    self.pointer += 1
                else:
                    return 0

            end = self.current.read(docs, freqs)
            if end == 0:
                self.current = None
            else:
                b = self.base
                for i in range(end):
                    docs[i] += b
                return end


    def seekTerm(self, term):
        self.term = term
        self.base = 0
        self.pointer = 0
        self.current = None


    def skipTo(self, target):
        while target > self.doc():
            if not self.next():
                return False

        return True


    def termDocsInt(self, i):
        if self.term is None:
            return None
        result = self.segTermDocs[i]
        if result is None:
            result = self.termDocsReader(self.readers[i])
            self.segTermDocs[i] = result
        result.seekTerm(self.term)
        return result


    def termDocsReader(self, reader):
        return reader.termDocs()



class SegmentsTermPositions(SegmentsTermDocs):
        
    def termDocsReader(self, reader):
        return reader.termPositions()


    def nextPosition(self):
        return self.current.nextPosition()
