# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""Examples of different kinds of query being run through an IndexSearcher.
"""
import os, sys
from cStringIO import StringIO

from lupy.analysis.simpleanalyzer import SimpleAnalyzer

from lupy.index.indexwriter import IndexWriter
from lupy.index.indexreader import IndexReader
from lupy.index.term import Term

from lupy.document.document import Document
from lupy.document import field

from lupy.search.indexsearcher import IndexSearcher
from lupy.search.termquery import TermQuery
from lupy.search.phrasequery import PhraseQuery
from lupy.search.booleanquery import BooleanQuery


class Index(object):

    def __init__(self, name, create = False, analyzer=SimpleAnalyzer):
        self.name = name
        self.analyzer = analyzer
        # Create the index if we need to. From here on we assume
        # that the index exists
        self.indexer = IndexWriter(self.name, self.analyzer(), create)
        # Remember the default merge factor
        self.mergeFactor = self.indexer.mergeFactor
        # Clean up
        self.indexer.close()
        self.indexer = self.searcher = None
        
    def index(self, **kw):
        """Add a document to the index.
        
        **kw contains the name and values of each Field in the
        Document that we are creating.

        If the key in **kw starts with '_' the field will be created
        as a Keyword. If it starts with '__', it is created as a
        stored Text field (e.g. tokenized and stored), otherwise it
        will be created as a Text field. The leading '_' are removed
        before field creation.

        Text fields will have their value tokenized before
        indexing. The value is not stored in the index.  This is the
        usual type of field that you need for plain text.

        Keyword fields will not have their value tokenized.  The value
        is stored in the index and is returned with search hits on the
        Document. If you wanted to store the path to a document along
        with each document, you would use a Keyword field. The path
        would not be tokenized and its value would be returned in the
        query results, so you could easily open and display the file.
        """
        self.setupIndexer()
        
        # create document
        d = Document()

        # TODO - Please find another way of defining fields
        # than magic field names!!!

        # add a file field containing the path to this file
        for key, value in kw.items():
            if key[:2] == '__':
                key = key[2:]
                # Tokenized and stored
                f = field.Text(key, value, True)
            elif key[0] == '_':
                # Not tokenized and stored
                key = key[1:]
                # keyword
                f = field.Keyword(key, value)
            else:
                # Tokenized and not stored
                f = field.Text(key, value, False)
            d.add(f)
        self.indexer.addDocument(d)

    def setupIndexer(self):
        if self.searcher is not None:
            self.searcher.close()
            self.searcher = None
        if self.indexer is None:
            a = self.analyzer()
            self.indexer = IndexWriter(self.name, a, create=False)
            self.indexer.mergeFactor = self.mergeFactor

    def setupSearcher(self):
        if self.indexer is not None:
            self.indexer.close()
            self.indexer = None
        if self.searcher is None:
            self.searcher = IndexSearcher(self.name)

    def delete(self, **kw):
        # Not very efficient for bulk deletes
        # Use deleteAll for bulk deletes
        self.setupSearcher()
        if len(kw) != 1:
            raise 'one and only one field for the moment'
        field, value = kw.items()[0]
        t = Term(field, value)
        self.searcher.reader.deleteTerm(t)
        
    def deleteAll(self, **kw):
        self.close()
        reader = IndexReader.open(self.name)
        if len(kw) != 1:
            raise 'one and only one field for the moment'
        field, values = kw.items()[0]
        for value in values:
            t = Term(field, value)
            reader.deleteTerm(t)
        # commit the deletes
        reader.close()

    def close(self):
        # Indexer and Searchers are different
        # and we have to open the right kind
        # for the operation we are performing.
        # The actual creation is done in the index and find
        # methods. Here we close whatever is open.
        if self.searcher is not None:
            self.searcher.close()
            self.searcher = None
        if self.indexer is not None:
            self.indexer.close()
            self.indexer = None

    def flush(self):
       """Flush outstanding indexes to disk.

       This makes sure we are searching the latest stuff.
       """
       if self.indexer is not None:
           self.indexer.flushRamSegments()

    def commit(self):
        self.close()

    def optimize(self):
        self.setupIndexer()
        self.indexer.optimize()

    def parse(self, field, qString):
        if qString.startswith('"'):
            qString = qString.strip('"')
            #qWords = qString.strip('"').split()
            qWords = self.tokenize(qString)
            return self.phraseSearch(field, qWords)
        else:
            qWords = self.tokenize(qString)
            if len(qWords) == 1:
                return self.termSearch(field, qWords[0])
            else:
                return self.boolSearch(field, qWords)

    def tokenize(self, qString):
        toks = []
        sio = StringIO(qString)
        a = self.analyzer()
        tokStream = a.tokenStream('ignore', sio)
        tok = tokStream.next()
        while tok:
            toks.append(tok.termText())
            tok = tokStream.next()
        return toks

    def find(self, qStr):
        # self.flush()
        # self.close()
        self.setupSearcher()
            
        fields = self.searcher.fieldNames()
        if not fields:
            return []
        all = [self.parse(field, qStr) for field in fields]
        if len(all) is 1:
            # simple case
            return self.searcher.searchQuery(all[0])
        
        q = BooleanQuery()
        for query in all:
            # OR all of the field queries
            q.add(query, False, False)
        hits = self.searcher.searchQuery(q)
        return hits

    def findInField(self, **kw):
        # eg index.findInField(text='flute')
        if len(kw) != 1:
            raise 'one and only one field for the moment'
        self.setupSearcher()
        field, query = kw.items()[0]
        q = self.parse(field, query)
        hits = self.searcher.searchQuery(q)
        return hits
    
    def termSearch(self, field, qStr):
        t = Term(field, qStr)
        q = TermQuery(t)
        return q

    def phraseSearch(self, field, words):
        """Find all docs containing the phrase I{"egg cream"}."""
        q = PhraseQuery()
        for word in words:
            t = Term(field, word)
            q.add(t)  
        return q
            
    def boolSearch(self, field, ands=[], ors=[], nots=[]):
        """Build a simple boolean query.

        Each word in B{ands} is equiv to +word
        Each word in B{ors} is equiv to word
        Each word in B{nots} is equiv to -word

        E.g. boolSearch(['spam'], ['eggs'], ['parrot', 'cheese']) is
        equiv to +spam eggs -parrot -cheese in Google/Lucene syntax.
        """
        q = BooleanQuery()

        for a in ands:
            t = Term(field, a)
            tq = TermQuery(t)
            q.add(tq, True, False)
            
        for a in ors:
            t = Term(field, a)
            tq = TermQuery(t)
            q.add(tq, False, False)
            
        for a in nots:
            t = Term(field, a)
            tq = TermQuery(t)
            q.add(tq, False, True)
        
        return q
            
    def printHits(self, hits):
        if len(hits) == 0:
            print 'Nothing found!'
        else:
            for i in range(len(hits)):
                print hits.doc(i), hits.score(i)

    def setMergeFactor(self, anInt):
        # Never less than 2
        if anInt >= 2:
            self.mergeFactor = anInt
        
          
if __name__ == "__main__":
    """One shot search.
    Opens and closes the index on every query.
    """
    
    # TODO Command line argument passing
    # TODO e.g.
    # TODO -d directory to store index in
    # TODO -i directory to recursively index
    
    import time
    tt = time.time()

    filedir = './aesop/'
    indexName = 'aesopind'

    if os.path.exists(indexName):
        for f in os.listdir(indexName):
            os.remove(os.path.join(indexName, f))
        # Remove results of previous runs
        os.rmdir(indexName)

    # Create a new Index
    index = Index(indexName, create = True)

    # Get the files
    files = os.listdir(filedir)
    for f in files:
        text = open(filedir + f).read()

        print 'indexing:', f
        id = f.split('.')[0]
        # id is a string
        # it is returned with this document
        # and returned with the hits. It is not
        # used by Lupy
        index.index(id, text)

    #index.optimize()
    index.commit()

    # re-open it for search
    index = Index(indexName)

    print 'fox:', index.find('fox')
    print 'python:', index.find('python')
    print 'fox stork', index.find('fox stork')
    print 'fox python', index.find('fox python')
    print '"the fox and the":', index.find('"the fox and the"')
    print '"the fox and python":', index.find('"the fox and python"')
    index.close()
    print 'Elapsed time:', time.time() - tt
