# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

__doc__ = """A Query that matches documents matching boolean combinations of
            other queries, typically L{TermQuery}s or L{PhraseQuery}s."""

import query
import booleanclause
import booleanscorer

class BooleanQuery(query.Query):
    __slots__ = ['clauses']
    

    def __init__(self):
        """Constructs an empty boolean query."""
        
        self.clauses = []


    def addClause(self, clause):
        """Adds a clause to a boolean clause."""
        self.clauses.append(clause)


    def add(self, query, required, prohibited):
        """Adds a clause to a boolean query.  Clauses may be:
        U{required} which means that documents which I{do not}
        match this sub-query will I{not} match the boolean query;
        U{prohibited} which means that documents which I{do}
        match this sub-query will i{not} match the boolean query; or
        U{neither, in which case matched documents are neither prohibited from
        nor required to match the sub-query.}
        
        It is an error to specify a clause as both B{required} and
        B{prohibited}."""
        
        self.clauses.append(booleanclause.BooleanClause(query,
                                                        required,
                                                        prohibited))


    def normalize(self, norm):
        for c in self.clauses:
            if not c.prohibited:
                c.query.normalize(norm)



    def preparer(self, reader):
        for c in self.clauses:
            c.query.prepare(reader)


    def scorer(self, reader):
        # optimize zero-term case
        if len(self.clauses) == 1:
            # just return term scorer
            c = self.clauses[0]
            if not c.prohibited:
                return c.query.scorer(reader)

        result = booleanscorer.BooleanScorer()

        for c in self.clauses:
            subScorer = c.query.scorer(reader)
            if subScorer is not None:
                result.add(subScorer, c.required, c.prohibited)
            elif c.required:
                return None

        return result
            

    def sumOfSquaredWeights(self, searcher):
        sum = 0.0
        
        for c in self.clauses:
            if not c.prohibited:
                # sum sub-query weights
                sum += c.query.sumOfSquaredWeights(searcher)
                
        return sum


    def toString(self, field):
        """Prints a user-readable version of this query"""

        buffer = ''

        for i in range(len(self.clauses)):
            c = self.clauses[i]
            if c.prohibited:
                buffer += '-'
            elif c.required:
                buffer += '+'

            subQuery = c.query
            if isinstance(subQuery, BooleanQuery):
                # wrap sub-bools in parens
                buffer += '('
                buffer += c.query.toString(field)
                buffer += ')'
            else:
                buffer += c.query.toString(field)

            if i != len(self.clauses) - 1:
                buffer += ' '

        return buffer


 
