# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

import scorer
import similarity
import hitcollector


class BooleanScorer(scorer.Scorer):
    
    __slots__ = ['bucketTable', 'coordFactors', 'currentDoc', 'maxCoord',
                 'nextMask', 'prohibitedMask', 'requiredMask', 'scorers']
    
    
    def __init__(self):
        self.bucketTable = BucketTable(self)
        self.coordFactors = None
        self.maxCoord = 1
        self.nextMask = 1
        self.prohibitedMask = 0
        self.requiredMask = 0
        self.scorers = None
        self.currentDoc = 0
        
        
    def add(self, scorer, required, prohibited):
        mask = 0
        if required or prohibited:
            if self.nextMask == 0:
                raise Exception, 'More than 32 required/prohibited clauses in a query.'
            mask = self.nextMask
            self.nextMask = self.nextMask << 1
        else:
            '???'
            mask = 0
            
        if not prohibited:
            self.maxCoord += 1
            
        if prohibited:
            # Update prohibited mask
            self.prohibitedMask |= mask
        elif required:
            # Update required mask
            self.requiredMask |= mask
            
        self.scorers = SubScorer(scorer, required, prohibited,
                                 self.bucketTable.newCollector(mask),
                                 self.scorers)
        
        
    def computeCoordFactors(self):
        self.coordFactors = []
        for i in range(self.maxCoord):
            self.coordFactors.append(similarity.coord(i, self.maxCoord))
            
            
    def score(self, results, maxDoc):
        if self.coordFactors is None:
            self.computeCoordFactors()
            
        while self.currentDoc < maxDoc:
            self.currentDoc = min(self.currentDoc + BucketTable.SIZE, maxDoc)
            t = self.scorers
            while t is not None:
                t.scorer.score(t.collector, self.currentDoc)
                t = t.next
            self.bucketTable.collectHits(results)
            
            
            
class BucketTable(object):
    """A simple hash table of document scores within a range."""

    SIZE = 1 << 10
    MASK = SIZE - 1
    
    __slots__ = ['buckets', 'first', 'scorer']
    
    def __init__(self, scor):
        self.scorer = scor
        
        self.buckets = [None] * self.SIZE
        # head of valid list
        self.first = None
        
        
    def collectHits(self, results):
        required = self.scorer.requiredMask
        prohibited = self.scorer.prohibitedMask
        coord = self.scorer.coordFactors
        
        bucket = self.first
        while bucket is not None:
            if (bucket.bits & prohibited) == 0 and (bucket.bits & required) == required:
                # if prohibited and required check out
                # add to results
                results.collect(bucket.doc, bucket.score * coord[bucket.coord])
            bucket = bucket.next
                
        # reset for next round
        self.first = None
    
        
    def newCollector(self, mask):
        return Collector(mask, self)
        
    def __len__(self):
        return self.SIZE
    
        
    
class Bucket(object):
    __slots__ = ['bits', 'coord', 'doc', 'next', 'score']
    
    def __init__(self):
        # tells if bucket is valid
        self.bits = 0
        self.coord = 0
        self.doc = -1
        
        
class Collector(hitcollector.HitCollector):
    __slots__ = ['bucketTable', 'mask']
    
    def __init__(self, mask, bucketTable):
        self.mask = mask
        self.bucketTable = bucketTable
        
        
    def collect(self, doc, score):
        table = self.bucketTable
        i = doc & BucketTable.MASK
        bucket = table.buckets[i]
        if bucket is None:
            bucket = Bucket()
            table.buckets[i] = bucket
            
        if bucket.doc != doc:
            # invalid doc
            # initialize fields
            bucket.doc = doc
            bucket.score = score
            bucket.bits = self.mask
            bucket.coord = 1
            
            # push onto valid list
            bucket.next = table.first
            table.first = bucket
        else:
            # valid bucket
            # increment score
            bucket.score += score
            # add bits in mask
            bucket.bits |= self.mask
            # increment coord
            bucket.coord += 1
            
    
    
class SubScorer(object):
    __slots__ = ['collector', 'next', 'prohibited', 'required', 'scorer']
    
    def __init__(self, scorer, required, prohibited, collector, next):
      self.scorer = scorer
      self.required = required
      self.prohibited = prohibited
      self.collector = collector
      self.next = next
    
    
    
    
