# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

__doc__ = """The abstract base class for search implementations.

          Subclasses implement search over a single index, over
          multiple indices, and over indices on remote servers."""


import similarity
import searcher, hits
import query as querymodule
import hitqueue
import hitcollector
import scoredoc
import topdocs

import lupy.index.indexreader as indexreader

class IndexSearcher(searcher.Searcher):
    __slots__ = ['reader']


    def __init__(self, dirOrPath):
        """Creates a searcher searching the provided index.
        """
        if type(dirOrPath) == type(''):
            # path
            self.reader = indexreader.IndexReader.open(dirOrPath)
        else:
            # dir
            self.reader = indexreader.IndexReader.openDir(dirOrPath)
        

    def setReader(self, r):
        self.reader = r
        

    def close(self):
        """Frees resources associated with this Searcher."""
        self.reader.close()


    def doc(self, i):
        """For use by L{HitCollector} implementations."""
        return self.reader.document(i)


    def docFreq(self, term):
        return self.reader.docFreq(term)


    def maxDoc(self):
        return self.reader.maxDoc()


    def searchQueryFilterHitColl(self, query, filter, results):
        """Lower-level search API.

        L{HitCollector#collect(int,float)} is called for every
        non-zero scoring document.

        Applications should only use this if they need <i>all</i> of the
        matching documents.  The high-level search API
        ({L{Searcher#search(Query)}) is usually more efficient, as it skips
        non-high-scoring hits.

        - B{query} to match documents
        - B{filter} if non-null, a bitset used to eliminate some documents
        - B{results} to receive hits
        """

        collector = results
        if filter is not None:
            bits = filter.bits(reader)
            collector = hitcollector.HitCollector()
            # set collect method for hitCollector
            # if bits.get(doc):
            #     results.collect(doc, score)

        scorer = query.Query.scorer(query, self, self.reader)
        if scorer is  None:
            return

        scorer.score(collector, self.reader.maxDoc())
            
    

    def searchQueryFilterInt(self, query, filter, nDocs):
        scorer = querymodule.Query.getScorer(query, self, self.reader)
        if scorer is None:
            return topdocs.TopDocs(0, [])

        if filter is not None:
            bits = filter.bits(reader)
        else:
            bits = None

        hq = hitqueue.HitQueue(nDocs)
        totalHits = [0]
        minScore = 0.0
                
        
        hc = hitcollector.HitCollector()
        # Load up the variables we will write to 
        # in _cbHitCollect
        hc.minScore = minScore
        
        def _cbHitCollect(doc, scr):
            if scr > 0.0 and (bits is None or bits.get(doc)):
                # ignore zeroed buckets and docs not in bits
                totalHits[0] += 1
                if scr >= hc.minScore:
                    # update hit queue
                    hq.put(scoredoc.ScoreDoc(doc, scr))
                    if hq.size() > nDocs:
                        # if hit queue overfull
                        # remove lowest in hit queue
                        hq.pop()
                        # reset minimum score
                        hc.minScore = hq.top().score
                
        hc.collect = _cbHitCollect
        scorer.score(hc, self.reader.maxDoc())

        scoreDocs = [0] * hq.size()
        for i in range(hq.size() - 1 , -1, -1):
            scoreDocs[i] = hq.pop()

        return topdocs.TopDocs(totalHits[0], scoreDocs)

    def fieldNames(self):
        # Experimental for auto queries
        return self.reader.fieldNames()
