# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

__doc__ = """Scorer for L{TermQuery}s."""

import scorer
import similarity
import sys

class TermScorer(scorer.Scorer):
    __slots__ = ['doc', 'docs', 'freqs', 'norms', 'pointer', 'pointerMax',
                 'scorerCache', 'termDocs', 'weight']

    SCORE_CACHE_SIZE = 32


    def __init__(self, td, n, w):
        self.termDocs = td
        self.norms = n
        self.weight = w
        # ????
        self.pointer = 0
        
        # buffered doc numbers
        self.docs = [0] * 128
        # buffered term freqs
        self.freqs = [0] * 128

        self.scoreCache = [0.0] * self.SCORE_CACHE_SIZE
        for i in range(len(self.scoreCache)):
            self.scoreCache[i] = similarity.tf(i) * self.weight
            
        # fill buffers
        self.pointerMax = self.termDocs.read(self.docs, self.freqs)

        if self.pointerMax != 0:
            self.doc = self.docs[0]
        else:
            # close stream
            self.termDocs.close()
            # set to sentinel value
            self.doc = sys.maxint


    def score(self, c, end):
        # cache doc in local
        d = self.doc

        # for docs in window
        while d < end:
            f = self.freqs[self.pointer]
            if f < self.SCORE_CACHE_SIZE:
                # cache hit
                score = self.scoreCache[f]
            else:
                # cache miss
                score = similarity.tf(f) * self.weight

            # normalize for field
            score *= similarity.normByte(self.norms[d])
            # collect score
            c.collect(d, score)

            self.pointer += 1
            if self.pointer == self.pointerMax:
                # refill buffers
                self.pointerMax = self.termDocs.read(self.docs, self.freqs)
                if self.pointerMax != 0:
                    self.pointer = 0
                else:
                    # close stream
                    self.termDocs.close()
                    # set to sentinel value
                    self.doc = sys.maxint
                    return
            d = self.docs[self.pointer]
        # flush cache
        self.doc = d
                    



        
        
        
        



        
        
