# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

from array import array
import weakref
import os, stat

from inputstream import InputStream
from outputstream import OutputStream
from directory import Directory


DIRECTORIES = weakref.WeakValueDictionary()


class FSDirectory(Directory):
    __slots__ = ['file', 'directory']

    # class methods
    
    def getDirectory(cls, path, gen):
        dir = DIRECTORIES.get(path, None)

        if dir is None:
            dir = FSDirectory(path, gen)
            DIRECTORIES[path]=dir
        elif gen is True:
            dir.create()

        return dir

    getDirectory = classmethod(getDirectory)

    # instance methods

    def __init__(self, path, gen):
        self.directory = path
        if gen is True:
            self.create()
        DIRECTORIES[path]=self

    def fpath(self, fname):
        return os.path.join(self.directory, fname)

    def create(self):
        path = self.directory
        if not os.path.exists(path):
            os.mkdir(path)
        try:
            files = os.listdir(path)
        except IOError:
            files = []
            
        for file in files:
            os.remove(os.path.join(path,file))

    def list(self):
        return os.listdir(self.directory)

    def fileExists(self, name):
        return os.path.exists(self.fpath(name))

    def fileModified(self, name, path=None):
        return os.stat(self.fpath(name))[stat.ST_MTIME]
    
    def fileLength(self, name):
        return os.stat(self.fpath(name))[stat.ST_SIZE]
    
    def deleteFile(self, name):
        os.remove(self.fpath(name))

    def renameFile(self, frm, to):
        if os.path.exists(self.fpath(to)):
            os.remove(self.fpath(to))

        os.rename(self.fpath(frm),self.fpath(to))

    def createFile(self, name):
        return FSOutputStream(self.fpath(name))

    def openFile(self, name):
        return FSInputStream(self.fpath(name))

    def close(self):
        del(DIRECTORIES[self.directory])

    def __str__(self):
        return 'FSDirectory:' + self.directory



class FSInputStream(InputStream):
    # __slots__ = ['file', 'length', 'isClone']


    def __init__(self, path):
        InputStream.__init__(self)
        
        self.file = Descriptor(path,'rb')
        self.file.position = 0
        self.isClone = False
        self.length = os.stat(path)[stat.ST_SIZE]


    def readInternal(self, b, offset, length):
        position = self.getFilePointer()
        if position != self.file.position:
            self.file.seek(position)
            self.file.position = position

        total = 0
        while total < length:
            bytes = self.file.read(length-total)

            lb = len(bytes)

            b[0:lb]=array('B',bytes)
            
            self.file.position += lb
            total += lb
            
          
    def seekInternal(self, position):
        """Random access method.
        """


    def close(self):
        if self.isClone is not True:
            self.file.close()



class FSOutputStream(OutputStream):
    __slots__ = ['file']


    def __init__(self, path):
        OutputStream.__init__(self)
        self.file = open(path,'wb')
        

    def flushBuffer(self, b, size):
        self.file.write(b[:size])
        

    def close(self):
        OutputStream.close(self)
        self.file.flush()
        self.file.close()
        

    def seek(self, pos):
        OutputStream.seek(self, pos)
        self.file.seek(pos)
        

    def length(self):
        return len(self.file)
    
    
    
class Descriptor(file):
    __slots__ = ['position']
    
    
    #def __init__(self, mode='rb'):
        #file.__init__(self, mode)
        #self.position = 0




