# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

# import time

from array import array

from inputstream import InputStream
from outputstream import OutputStream
from directory import Directory


class RAMDirectory(Directory):
    __slots__ = ['files']

    
    def __init__(self):
        self.files = {}

    
    def list(self):
        return self.files.keys()


    def fileExists(self, name):
        return (self.files.get(name, None) is not None)


    def fileModified(self, name):
        file = self.files[name]
        return file.lastModified
    

    def fileLength(self, name):
        file=self.files[name]
        return len(file)
    

    def deleteFile(self, name):
        del(self.files[name])


    def renameFile(self, name, newName):
        file = self.files[name]
        del(self.files[name])
        self.files[newName]=file


    def createFile(self, name):
        file = RAMFile()
        self.files[name]=file
        return RAMOutputStream(file)


    def openFile(self, name):
        file = self.files[name]
        return RAMInputStream(file)


    def makeLock(self, name):
        """TBC"""


    def close(self):
        """Do nothing"""



class RAMInputStream(InputStream):

    #__slots__ = ['file', 'pointer', 'length', 'isClone']

    def __init__(self, f):
        InputStream.__init__(self)
        
        self.pointer = 0
        self.file = f
        self.length = self.file.length


    def readInternal(self, dest, destOffset, length):

        remainder = length
        start = self.pointer
        
        while remainder != 0:
            bufferNumber = start/InputStream.BUFFER_SIZE
            bufferOffset = start%InputStream.BUFFER_SIZE
            bytesInBuffer = InputStream.BUFFER_SIZE - bufferOffset
            
            if bytesInBuffer >= remainder:
                bytesToCopy = remainder
            else:
                bytesToCopy = bytesInBuffer

            self.buffer = self.file.buffers[bufferNumber]
            
            bo = bufferOffset
            do = destOffset
            dest[do:do+bytesToCopy] = self.buffer[bo:bo+bytesToCopy]

            destOffset += bytesToCopy
            start += bytesToCopy
            remainder -= bytesToCopy

        self.pointer += length


    def close(self):
        """Do nothing"""


    def seekInternal(self, pos):
        self.pointer = pos



class RAMOutputStream(OutputStream):
    __slots__ = ['pointer', 'file']

    def __init__(self, f):
        OutputStream.__init__(self)
        
        self.file = f
        self.pointer = 0


    def flushBuffer(self, src, length):
        bufferNumber = self.pointer/OutputStream.BUFFER_SIZE
        bufferOffset = self.pointer%OutputStream.BUFFER_SIZE
        bytesInBuffer = OutputStream.BUFFER_SIZE - bufferOffset

        if bytesInBuffer >= length:
            bytesToCopy = length
        else:
            bytesToCopy = bytesInBuffer

        if bufferNumber == len(self.file.buffers):
            self.file.buffers.append(array('B',[0x00]*OutputStream.BUFFER_SIZE))

        buffer = self.file.buffers[bufferNumber]
        buffer[bufferOffset:bufferOffset + bytesToCopy] = src[0:bytesToCopy]
        
        if bytesToCopy < length:
            srcOffset = bytesToCopy
            bytesToCopy = length - bytesToCopy
            bufferNumber += 1
            
            if bufferNumber == self.file.buffers.size():
                self.file.buffers.append(array('B',[0x00]*OutputStream.BUFFER_SIZE))

            buffer = self.file.buffers[bufferNumber]
            buffer[0:bytesToCopy]=src[srcOffset:srcOffset+bytesToCopy]
        self.pointer += length
        if self.pointer > self.file.length:
            self.file.length = self.pointer

        # self.file.lastModified = time.time()


    def seek(self, pos):
        OutputStream.seek(self, pos)
        self.pointer = pos
        

    def length(self):
        return self.file.length
        


class RAMFile(object):
    # __slots__ = ['buffers', 'length', 'lastModified']

    def __init__(self):
        self.length = 0
        self.buffers = []
        # self.lastModified = time.time()

