# Carlo B. Bifulco
# Mostly based and taken from Lupy's main indexer module 
# See following lines for Lupy's Copyright:
# This module is part of the Lupy project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

from __future__ import generators
import os, sys

from lupy.analysis.simpleanalyzer import SimpleAnalyzer
from lupy.index.indexwriter import IndexWriter
from lupy.index.indexreader import IndexReader
from lupy.index.term import Term
from lupy.document.document import Document
import lupy.document.field as field
from lupy.search.indexsearcher import IndexSearcher
from lupy.search.termquery import TermQuery
from lupy.search.phrasequery import PhraseQuery
from lupy.search.booleanquery import BooleanQuery
from shutil import rmtree
import sharedGlobals



oK=sharedGlobals.Share()

def slashNormalizer(pathName):
    if os.name!="posix":
        return pathName.replace("\\","/")
    else:
        return pathName



def enumerate(sequence):
    index = 0
    for item in sequence:
        yield index, item
        index += 1



class Parser:
    def __init__(self):
        self.error=[]
        self.nonParsableFiles=[]
    def parse(self):
        raise "Not implemented"        

class TextParser(Parser):
    def __init__(self):
        Parser.__init__(self)
    def parse(self,fileName):
        fh=file(fileName,"r")
        text=fh.read()
        fh.close()
        return text
    


class Indexer(object):
    def __init__(self, o_Parser, rootPath=oK.testDir):
        """Create an indexer, writing and index to the directory B{path}.
        if index does nor exist _> new
        else just update"""
        assert os.path.isdir(rootPath),"%s is not an existing directory" %rootPath
        self.rootPath=slashNormalizer(os.path.abspath(rootPath))
        self.indexPath=slashNormalizer(os.path.abspath(os.path.join(rootPath,oK.indexDirName)))
        
        self.open=False
        self.o_Parser=o_Parser
        try:
            self._updateOldIndex()
        except:
            self._createNewIndex()
      

    def __repr__(self):
        print "%s is the directory containing the index data structure" %self.indexPath
        print "%s is the directory being actually indexed" %self.rootPath
        if self.open:
            print "The index is currently open"
        else:
            print "The index is closed"
        print "%s is the parser" %self.o_Parser
        return " "
    
    def _createNewIndex(self):
        create=True
        self.indexer = IndexWriter(self.indexPath, SimpleAnalyzer(), create)
        #print "created index new index"
        self.open=True
        
    def _updateOldIndex(self):
        """ raises exception if directory does not exist"""
        create=False
        self.indexer = IndexWriter(self.indexPath, SimpleAnalyzer(), create)
        #print "updating index", self.indexPath
        self.open=True
        
    def _checkDir(self):
        if not os.path.isdir(self.indexPath):
            #os.mkdir(self.indexPath)
            return False
        else:
            return True
        
    def _getFiles(self, filesDir):
        def getFileNames (a,directory,files):
            for fileName in files:
                fullName=os.path.abspath(os.path.join(directory,fileName))
                if os.path.isfile(fullName):
                    self._allFileNames.append(fullName)
        os.path.walk(filesDir,getFileNames,None)
        
    def _addDoc(self, fname):
        assert self.open, "INDEX NOT OPEN!"
        assert os.path.isfile(fname), "%s is not an existing file" %fname
        fname=slashNormalizer(os.path.abspath(fname))
        
        """Add a document to the index"""
        # create document
        d = Document()
        # add a file field containing the path to this file
        f = field.Keyword('file',fname)
        d.add(f)
        parsedText=self.o_Parser.parse(fname)
        if parsedText==False:
            return False
        else:
            f = field.Text('text', parsedText, False)
            d.add(f)
            self.indexer.addDocument(d)
            return True

    def indexDir(self):
        if not self.open:
            self.reOpenIndex()
        self._allFileNames=[]
        self._getFiles(self.rootPath) # -> self._allFileNames
        self._allFileNames=[slashNormalizer(i) for i in self._allFileNames if i.find("indexDir") ==-1]
        self.indexedFiles=self.getAllIndexed()
        #print indexedFiles
        toBeIndexed=[i for i in self._allFileNames if i not in self.indexedFiles]
        numberOfFiles=len(self._allFileNames)
        counter=0
        for fileName in toBeIndexed:
            #print fileName
            self._addDoc(fileName)
            counter+=1
            #print "%s out of %s" %(counter, numberOfFiles)
        for fileName in self.indexedFiles:
            if fileName not in self._allFileNames:
                #print "remove", fileName
                if not self.remove(fileName):
                    print "FAILED TO REMOVE %s" %fileName

    def index(self, filePath):
        return self._addDoc(filePath)

    def reset(self):
        rmtree(self.indexPath)
        self._createNewIndex()

    def reOpenIndex(self):
        assert os.path.isdir(self.indexPath),"NO INDEX PRESENT TO OPEN"
        assert not self.open, "INDEX ALREADY OPEN!"
        self._updateOldIndex()
        self.open=1

    def optimize(self):
        if self.open:
            #print "optimizing"
            self.indexer.optimize()
        else:
            #print "optimizing"
            self.reOpenIndex()
            self.indexer.optimize()
            self.close()

    def close(self):
        assert self.open, "INDEX IS NOT OPEN"
        self.indexer.flushRamSegments()
        self.indexer.close()
        self.indexer=None
        self.open=False
        
    def getAllIndexed(self):
        if self.open:
            self.close()
            fileNames=self._getAllIndexed()
            self.reOpenIndex()
            return fileNames
        else:
            return self._getAllIndexed()

    def _getAllIndexed(self):
        o_Reader=IndexReader.open(self.indexPath)
        fileNames=[]
        for n in range(o_Reader.maxDoc()):
            if not o_Reader.isDeleted(n):
                doc=o_Reader.document(n)
                fileName=slashNormalizer(doc.getField("file").stringValue())
                #print "%s has newer been deleted" % fileName
                fileNames.append(fileName)
            else:
                pass
                #print "%s was deleted" % str(n)
        o_Reader.close()
        return fileNames
        
        
    def remove(self,fileName):
        if self.open:
            self.close()
        fileName=slashNormalizer(os.path.abspath(fileName))
        #print fileName,self.indexPath
        oRemover = IndexReader.open(self.indexPath)
        t = Term("file", fileName)
        result=oRemover.deleteTerm(t)
        #print result
        oRemover.close()
        if result!=0:
            #print "SUCCESS"
            pass
        return result





class Searcher:
    def __init__(self,rootPath=oK.testDir):
        """
        #>>> s=Searcher()
        #>>> s._boolSplit("Carlo  filippo OR CANE")
        #(['carlo', 'filippo'], ['cane'], [])
        #>>> s._boolSplit("carlo  filippo")
        #(['carlo', 'filippo'], [], [])
        #>>> s._boolSplit("Carlo  filippo NOT cane")
        #(['carlo', 'filippo'], [], ['cane'])
        #>>>
        """
        self.indexPath=slashNormalizer(os.path.abspath(os.path.join(rootPath,oK.indexDirName)))
        assert os.path.isdir(self.indexPath), "Indexed not found in %s" %self.indexPath
        self.oIndex=IndexSearcher(self.indexPath)

    def search(self,text):
        query=self._getQueryType(text)(text)
        results=self._runQuery(query)
        return [result.get("file") for result in results]

    def close(self):
        self.oIndex.close()

    def _getQueryType(self,text):
            """
            getQ("jhkjh KJh NOT kj AND kjh kjh khasasa asasas asa")
            string -> (['kjh'], [], ['kj'])
            where ([ANDs],[ORs], [NOTs])"""
            #print "querytype"
            words=text.split()
            if len(words)==1:
                return self._termSearch
            if text[0] == '"' and text[-1]=='"':
                print "PHRASE"
                return self._phraseSearch
            return self._boolSearch
    def _termSearch(self,qStr):
        #print "TERMSEARCH"
        qStr=qStr.lower()
        t = Term('text', qStr)
        q = TermQuery(t)
        return q

    def _phraseSearch(self,qStr):
        """Find all docs containing the phrase I{"egg cream"}."""
        qStr=qStr.lower()[1:-1] # '"xxxXXX"' -> 'xxxxxx'
        parts = qStr.split()
        q = PhraseQuery()
        for p in parts:
            t = Term('text', p)
            q.add(t)
        return q

    def _boolSplit(self,text):
        ands=[];ors=[];nots=[]
        words=text.split()
        for wordTuple in enumerate(words):
            wordString=wordTuple[1]
            wordIndex=wordTuple[0]
            #print "current", wordString
            if wordString in ["NOT", "OR", "AND"]: continue
            if wordIndex==0:
                if words[1]=="OR":
                    ors.append(wordString)
                else:
                    ands.append(wordString)
                continue
            previousWord=words[wordIndex-1]
            #print "prev", previousWord,"curr", wordString
            wordString=wordString.lower()
            if previousWord=="NOT":
               nots.append(wordString)
            if previousWord=="AND" or previousWord not in ["OR","NOT"]:
                ands.append(wordString)
            if previousWord=="OR":
                ors.append(wordString)
        ands=[i.lower() for i in ands]
        ors=[i.lower() for i in ors]
        nots=[i.lower() for i in nots]
        return ands, ors, nots
    
    def _boolSearch(self,text):
        """Build a simple boolean query.
        each word in B{ands} is equiv to +word
        each word in B{ors} is equiv to word
        each word in B{nots} is equiv to -word

        e.g. boolSearch(['spam'], ['eggs'], ['parrot', 'cheese'])
        is equiv to +spam eggs -parrot -cheese in Google/Lucene syntax"""
        ands,ors,nots= self._boolSplit(text)
        q = BooleanQuery()
        for a in ands:
            t = Term('text', a)
            tq = TermQuery(t)
            q.add(tq, True, False)
        for a in ors:
            t = Term('text', a)
            tq = TermQuery(t)
            q.add(tq, False, False)
        for a in nots:
            t = Term('text', a)
            tq = TermQuery(t)
            q.add(tq, False, True)
        #print ands,ors,nots
        return q
    def _runQuery(self, q):
        """The run a query through a searcher and return the hits"""
        print 'Query:', q.toString('text')
        hits = self.oIndex.searchQuery(q)
        return hits


def _test():
    import doctest, indexer
    return doctest.testmod(indexer)

__test__={"Indexer":"""
>>> import shutil
>>> shutil.rmtree("./testDir")
>>> shutil.copytree("./testDirCopy", "./testDir")
>>> oI=Indexer(TextParser(),"./testDir/testText")
>>> oI.indexDir()
>>> os.rename("./testDir/testText/nested1", "./testDir/testText/nested2")
>>> oI.indexDir()
>>> os.rename("./testDir/testText/nested2", "./testDir/testText/nested1")
>>> oI.indexDir()
>>> os.rename("./testDir/testText/nested1", "./testDir/testText/nested2")
>>> oI.indexDir()
>>> os.rename("./testDir/testText/nested2", "./testDir/testText/nested1")
>>> oI.indexDir()
>>> oI.optimize()
>>> os.rename("./testDir/testText/nested1", "./testDir/testText/nested2")
>>> oI.indexDir()
"""}


if __name__ =="__main__":
    _test()
